#include "bitmap.h"

#include <vector>

const int numFullPics = 43;
const int numPreRidingPics = 40;
const int numMinorPics = 9;
const int numCharacters = 165;

class Character;

class Character
{
public:

	Character(void);
	~Character(void);

	Bitmap pics[numFullPics];
	
	enum ChracterType {
		FULL,
		NO_RIDING,
		MINOR,
		STATIC_NPC,
		ANIMATED,
		RAMUH,
		GUARD_RIDING,
		STATIC_16,
		STATIC_32,
	} type;


	void changePalette(BYTE);
	void clear(void);
	void copyImage(int, int);
	void createBitmaps(void);
	void flipImage(int);
	void load(BYTE* ptr);
	void load(int, BYTE);
	void save(int, BYTE);
	void save(int handle);
	void setImage(Bitmap& bmp);
	void setPixel(BYTE, int, int, BYTE);
	void setTransparency(BYTE);
	void updatePalette(void);
	int getTileHeight() const;
	int getTileWidth() const;
	int getNumPics() const;
	bool canCopyFrom(const Character& other, int sprite) const;

	//Inline Functions
	inline BYTE getPalNum(void)
	{
		return palette;
	}

	int getPixelHeight() const { return 8 * getTileHeight();  }
	int getPixelWidth() const { return 8 * getTileWidth(); }
	int getNumTiles() const { return size / 0x20; }
	

private:
	const USHORT *Character::getTileOffsets() const;

	BYTE palette;
	int tileOffset;
	bool fixedPalette;
	int size;
	std::vector<BYTE> tiles;
};

bool loadCharacters(char*);
void loadCharPals(int);
void loadCharTiles(int);
bool saveCharacters(char *);
void saveCharPals(int);

extern Character character[numCharacters];
extern RGBTRIPLE charPals[32][16];

