#include "stdafx.h"
#include "resource.h"

#include "ff3se.h"
#include "tab1.h"
#include "monster.h"
#include "monsters.h"
#include "options.h"
#include "tab3.h"
#include "portrait.h"
#include "combatback.h"
#include "changeback.h"
#include "grid.h"
#include "tab4.h"
#include "character.h"

#define IDR_FILE_RECENT1 23457
#define IDR_FILE_RECENT2 23458
#define IDR_FILE_RECENT3 23459
#define IDR_FILE_RECENT4 23460

HINSTANCE hInst;
HWND hWnd;
HWND hMainTab;
HDC screenHdc;
HCURSOR brushCursor;
HCURSOR eraseCursor;
HCURSOR grabCursor;
HWND hDlg = NULL;			//Global handle for all dialog boxes, so they can work with isDialogMessage

HBITMAP backBMP, back2BMP;
Bitmap backBmp, back2Bmp;

bool headered = true;

int tabSelected = 0;

char ff3seTitle[80];
char ROMname[1024] = "";

void cleanup(void)
{
	monsters.cleanup();
	changeBackCleanup();

	DeleteObject(backBMP);
	DeleteObject(back2BMP);

	ReleaseDC(hWnd, screenHdc);

	DestroyWindow(hMainTab);
	DestroyWindow(hWnd);
}

/*************************************/
//Adds the recent file list to the File menu
void addRecentFileList(void)
{
	HMENU hMenu;
	HMENU hFile;
	MENUITEMINFO mii;
	char filename[256];
	int i;

	if (options.getNumRecentFiles() == 0)
		return;

	hMenu = GetMenu(hWnd);
	hFile = GetSubMenu(hMenu, 0);

	//Delete the current recent file list
	DeleteMenu(hFile, 4, MF_BYPOSITION);
	DeleteMenu(hFile, IDR_FILE_RECENT1, MF_BYCOMMAND);
	DeleteMenu(hFile, IDR_FILE_RECENT2, MF_BYCOMMAND);
	DeleteMenu(hFile, IDR_FILE_RECENT3, MF_BYCOMMAND);
	DeleteMenu(hFile, IDR_FILE_RECENT4, MF_BYCOMMAND);

	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_TYPE;
	mii.fType = MFT_SEPARATOR;
	mii.fState = MFS_ENABLED;
	mii.wID = 23456;
	mii.hSubMenu = NULL;
	mii.hbmpChecked = NULL;
	mii.hbmpUnchecked = NULL;
	mii.dwItemData = 0;
	mii.dwTypeData = filename;
	mii.cch = 256;

	InsertMenuItem(hFile, 4, true, &mii);

	for (i = 0; i < options.getNumRecentFiles(); i++)
	{
		sprintf(filename, "&%d %s", i + 1, options.getFullRecentFile(i));
		AppendMenu(hFile, MF_STRING, IDR_FILE_RECENT1 + i, filename);
	}

	DrawMenuBar(hWnd);
}

/*************************************/
//Converrts a Windows' COLORREF to and RGBTRIPLE
RGBTRIPLE COLORREFtoRGB(long color)
{
	RGBTRIPLE r;

	r.rgbtRed = color & 0xFF;
	r.rgbtGreen = (color & 0xFF00) >> 8;
	r.rgbtBlue = (color & 0xFF0000) >> 16;
	return r;
}

/*************************************/
//Checks if data has been changed.
//If it has checks if you want to save
//If cancel is selected, or yes is selected, and it can't save, then 1 is returned else 0
BYTE confirmClose(void)
{
	int result;

	if (tab1.hasChanged() || tab3.hasChanged() || tab4.hasChanged())
	{
		result = MessageBox(NULL, "Save changes to the ROM?", ff3seTitle, MB_YESNOCANCEL | MB_ICONQUESTION | MB_TASKMODAL);
		if (result == IDYES)
		{
			if (!saveROM())
				return 1;
		}
		else if (result == IDCANCEL)
			return 1;
	}
	return 0;
}

/*************************************/
//Reads data from the file handle passed in and decompresses it into data
void decompress(int handle, BYTE *data)
{
	unsigned char buf[8192];
	unsigned char buf2[2048];
	unsigned char n, x, b;
	int size = 0, w, num, i;
	long bpos, bpos2;
	int finalCount = 0;
	BYTE *pTemp = new BYTE[1048576];
	BYTE *pTempPtr = pTemp;

	ZeroMemory(buf2, 2048);
	ZeroMemory(buf, 8192);

	_read(handle, &size, 2);

	size = (size - 2) & 0xFFFF;

	_read(handle, &buf, size);

	bpos = 0; bpos2 = 2014;

	do
	{
		n = buf[bpos];
		bpos++;

		for (x = 0; x < 8; x++)
		{
			if (((n >> x) & 1) == 1)
			{
				b = buf[bpos];
				bpos++;
				*(pTemp++) = b;
				finalCount++;
				buf2[bpos2 & 2047] = b;
				bpos2++;
			}
			else
			{
				w = buf[bpos] + (buf[bpos + 1] << 8);
				bpos += 2;
				num = (w >> 11) + 3;
				w = w & 2047;

				for (i = 0; i < num; i++)
				{
					b = buf2[(w + i) & 2047];
					*(pTemp++) = b;
					finalCount++;
					buf2[bpos2 & 2047] = b;
					bpos2++;
				}
			}
			if (bpos >= size)
				x = 8;
		}
	} while (bpos < size);

	memcpy(data, pTempPtr, finalCount);
	SAFE_DELETE(pTempPtr);
}

/*************************************/
//Converts an RGBTRIPLE to an integer equivalent used by the ROM
void getIntPal(RGBTRIPLE color, BYTE *c1, BYTE *c2)
{
	*c1 = color.rgbtRed / 8;
	*c1 += ((color.rgbtGreen / 8) & 0x07) << 5;
	*c2 = ((color.rgbtGreen / 8) & 0x18) >> 3;
	*c2 += (color.rgbtBlue / 8) << 2;
}

/*************************************/
//Gets an RGBTRIPLE from 2 bytes in ROM format
RGBTRIPLE getRGBTRIPLE(BYTE c1, BYTE c2)
{
	RGBTRIPLE col;

	col.rgbtRed = (c1 & 0x1F) * 8;
	col.rgbtGreen = ((c1 & 0xE0) >> 5) + ((c2 & 0x03) << 3);
	col.rgbtGreen *= 8;
	col.rgbtBlue = (c2 & 0x7C) << 1;

	return col;
}

/*************************************/
//Returns a pointer to the currently selected tab
Tab *getSelectedTab(void)
{
	switch (tabSelected)
	{
	case 0:
	case 1:
		return &tab1;
	case 2:
		return &tab3;
	case 3:
		return &tab4;
	}

	return NULL;		//This should never be reached
}

/*************************************/
void initProgram(void)
{
	screenHdc = GetDC(hWnd);

	brushCursor = (HCURSOR)LoadImage(hInst, MAKEINTRESOURCE(IDC_CURSOR_BRUSH), IMAGE_CURSOR,
		32, 32, LR_DEFAULTCOLOR);
	eraseCursor = (HCURSOR)LoadImage(hInst, MAKEINTRESOURCE(IDC_CURSOR_ERASE), IMAGE_CURSOR,
		32, 32, LR_DEFAULTCOLOR);
	grabCursor = (HCURSOR)LoadImage(hInst, MAKEINTRESOURCE(IDC_CURSOR_GRAB), IMAGE_CURSOR,
		32, 32, LR_DEFAULTCOLOR);

	addRecentFileList();
}

/*************************************/
int InitTab(void)
{
	TCITEM tab;

	InitCommonControls();

	hMainTab = CreateWindow(WC_TABCONTROL, "", WS_CHILD | WS_CLIPSIBLINGS,
		0, 0, 680, 485, hWnd, NULL, hInst, NULL);

	if (hMainTab == NULL)
	{
		MessageBox(NULL, "Error initliazing program", ff3seTitle, MB_ICONERROR | MB_OK | MB_TASKMODAL);
		return 1;
	}

	tab.mask = TCIF_TEXT;
	tab.pszText = "Monsters";
	TabCtrl_InsertItem(hMainTab, 0, &tab);

	tab.mask = TCIF_TEXT;
	tab.pszText = "Espers";
	TabCtrl_InsertItem(hMainTab, 1, &tab);

	tab.mask = TCIF_TEXT;
	tab.pszText = "Portraits";
	TabCtrl_InsertItem(hMainTab, 2, &tab);

	tab.mask = TCIF_TEXT;
	tab.pszText = "Characters";
	TabCtrl_InsertItem(hMainTab, 3, &tab);

	/*    tab.mask = TCIF_TEXT;
		tab.pszText = "Menu Font";
		TabCtrl_InsertItem(hTab, 4, &tab);

		tab.mask = TCIF_TEXT;
		tab.pszText = "Cursors";
		TabCtrl_InsertItem(hTab, 5, &tab);*/

	tab1.create();
	tab3.create();
	tab4.create();

	return 0;
}

/*************************************/
int initWindow(void)
{
	WNDCLASS window;
	int x, y;
	RECT r;

	window.cbClsExtra = 0;
	window.cbWndExtra = 0;
	window.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	window.hCursor = LoadCursor(NULL, IDC_ARROW);
	window.hIcon = (HICON)LoadImage(hInst, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON, 32, 32, 0);
	window.hInstance = hInst;
	window.lpfnWndProc = (WNDPROC)WndProc;
	window.lpszClassName = "Main Window";
	window.lpszMenuName = NULL;
	window.style = 0;

	RegisterClass(&window);

	GetWindowRect(GetDesktopWindow(), &r);

	x = options.getXPos();
	if ((x < 0) || (x > r.right))
		x = 50;
	y = options.getYPos();
	if ((y < 0) || (y > r.bottom))
		y = 50;

	hWnd = CreateWindow("Main Window", ff3seTitle, WS_BORDER | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE,
		x, y, 685, 500, NULL, LoadMenu(hInst, MAKEINTRESOURCE(IDR_MENU)), hInst, NULL);

	if (hWnd == NULL)
	{
		MessageBox(NULL, "Error initliazing program", ff3seTitle, MB_ICONERROR | MB_OK | MB_TASKMODAL);
		return 1;
	}

	return 0;
}

/*************************************/
//Makes an RGBTRIPLE variable from it's 3 colors component bytes
RGBTRIPLE makeRGBTRIPLE(BYTE r, BYTE g, BYTE b)
{
	RGBTRIPLE temp;

	temp.rgbtBlue = b;
	temp.rgbtGreen = g;
	temp.rgbtRed = r;

	return temp;
}

/*************************************/
//Opens  a ROM and loads all the data into memory
//returns 0 if succesful, 1 if user cancelled, and 2 if an error occured
//If num > 0 then recenct file #num is opened, else a dialog box appears to get the filename
BYTE openROM(BYTE num)
{
	OPENFILENAME of;
	char filename[1024];
	HMENU hMenu;
	int handle;
	BYTE c;
	HWND hLoad, hText;
	RECT r;

	if (confirmClose() == 1)
		return 0;

	if (num < 4)
	{
		strcpy(filename, options.getFullRecentFile(num));
	}
	else
	{
		ZeroMemory(filename, 1024);

		char title[80] = "Select location of Final Fantasy III ROM";

		of.Flags = OFN_EXPLORER | OFN_HIDEREADONLY | OFN_FILEMUSTEXIST;
		of.hInstance = hInst;
		of.hwndOwner = hWnd;
		of.lCustData = 0;
		of.lpfnHook = NULL;
		of.lpstrCustomFilter = NULL;
		of.lpstrDefExt = NULL;
		of.lpstrFile = filename;
		of.lpstrFileTitle = NULL;
		of.lpstrFilter = "ROM file (*.smc)\000*.smc;\000All files (*.*)\000*.*\000\000";
		of.lpstrInitialDir = NULL;
		of.lpstrTitle = title;
		of.lpTemplateName = NULL;
		of.lStructSize = sizeof(OPENFILENAME);
		of.nFileExtension = 0;
		of.nFileOffset = 0;
		of.nFilterIndex = 0;
		of.nMaxCustFilter = 0;
		of.nMaxFile = 1024;
		of.nMaxFileTitle = 0;

		if (!GetOpenFileName(&of))
			return 1;
	}


	hLoad = CreateDialog(hInst, MAKEINTRESOURCE(IDD_LOADING), hWnd, NULL);
	SetWindowText(hLoad, ff3seTitle);
	hText = GetDlgItem(hLoad, IDC_TEXT_LOAD);
	SetWindowText(hText, "Loading Final Fantasy III ROM ...");
	GetWindowRect(hWnd, &r);
	SetWindowPos(hLoad, HWND_NOTOPMOST, r.left + 150, r.top + 100, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);
	UpdateWindow(hLoad);

	handle = _open(filename, O_RDWR | O_BINARY);

	if (handle == -1)
	{
		MessageBox(NULL, "Error loading ROM", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 2;
	}

	//Check to see if it's a valid Final Fantasy III ROM
	_lseek(handle, 0x220, SEEK_SET);

	_read(handle, &c, 1);
	if (c == 0xA2)
	{
		headered = true;
	}
	else
	{
		_lseek(handle, 0x20, SEEK_SET);
		_read(handle, &c, 1);
		if (c == 0xA2)
		{
			headered = false;
		}
		else
		{
			if (MessageBox(NULL, "This file does not appear to be a valid Final Fantasy III ROM. Do you want to load it anyway?",
				ff3seTitle, MB_YESNO | MB_ICONWARNING | MB_TASKMODAL) == IDNO)
			{
				DestroyWindow(hLoad);
				_close(handle);

				return 1;
			}
		}
	}
	_close(handle);

	if (!monsters.loadMonsters(filename))
	{
		MessageBox(NULL, "Error loading ROM", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 2;
	}

	if (!loadPortraits(filename))
	{
		MessageBox(NULL, "Error loading ROM", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 2;
	}

	if (!loadCharacters(filename))
	{
		MessageBox(NULL, "Error loading ROM", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 2;
	}

	DestroyWindow(hLoad);

	if (ROMname[0] == 0)			//The first time a ROM is opened, enable Save and Image
	{						//change the background, and show tab selected
		strcpy(ROMname, filename);

		setBackground();

		ShowWindow(hMainTab, true);

		tab1.initTab();
		tab3.initTab();
		tab4.initTab();

		if ((tabSelected < 0) || (tabSelected > 3))
			tabSelected = 0;

		switchToTab(tabSelected);
		TabCtrl_SetCurSel(hMainTab, tabSelected);

		hMenu = GetMenu(hWnd);
		EnableMenuItem(hMenu, IDR_FILE_SAVE, MF_ENABLED);
		EnableMenuItem(hMenu, IDR_OPTIONS_CHANGEBACK, MF_ENABLED);
		EnableMenuItem(hMenu, 1, MF_BYPOSITION | MF_ENABLED);
	}

	strcpy(ROMname, filename);

	getSelectedTab()->newROM();

	options.addRecentFile(filename);
	addRecentFileList();

	return 0;
}

/*************************************/
//Save the changed ROM to disk
bool saveROM(void)
{
	bool success;
	HWND hLoad, hText;
	RECT r;

	hLoad = CreateDialog(hInst, MAKEINTRESOURCE(IDD_LOADING), hWnd, NULL);
	SetWindowText(hLoad, ff3seTitle);
	hText = GetDlgItem(hLoad, IDC_TEXT_LOAD);
	SetWindowText(hText, "Saving Final Fantasy III ROM ...");
	GetWindowRect(hWnd, &r);
	SetWindowPos(hLoad, HWND_NOTOPMOST, r.left + 150, r.top + 130, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);
	UpdateWindow(hLoad);

	success = tab1.save(ROMname);
	if (!success)
	{
		DestroyWindow(hLoad);
		return false;
	}

	success = tab3.save(ROMname);
	if (!success)
	{
		DestroyWindow(hLoad);
		return false;
	}

	success = tab4.save(ROMname);
	if (!success)
	{
		DestroyWindow(hLoad);
		return false;
	}

	DestroyWindow(hLoad);

	MessageBox(NULL, "ROM saved", ff3seTitle, MB_OK | MB_TASKMODAL);

	return true;
}

/*************************************/
//Shows The tab tha tis currently selected
void switchToTab(BYTE num)
{
	tabSelected = num;

	if (num > 1)
		tab1.hide();
	if (num != 2)
		tab3.hide();
	if (num != 3)
		tab4.hide();

	getSelectedTab()->show();

	if (num < 2)		//If Monster or Esper tab
		tab1.switchType();

}

/*************************************/
//return the character equivalent of tile n in the ROM
char tileChar(BYTE n)
{
	if (n == 1)
		return ' ';
	else if (n == 5)
		return ' ';
	else if ((n >= 30) && (n <= 34))
		return ' ';
	else if ((n >= 128) && (n <= 153))
		return (n - 63);
	else if ((n >= 154) && (n <= 179))
		return (n - 57);
	else if ((n >= 180) && (n <= 189))
		return (n - 132);
	else if (n == 190)
		return '!';
	else if (n == 191)
		return '?';
	else if (n == 192)
		return'/';
	else if (n == 194)
		return'"';
	else if (n == 195)
		return'\'';
	else if (n == 196)
		return'-';
	else if (n == 197)
		return'.';
	else if (n == 198)
		return',';
	else if (n == 199)
		return' '; //..
	else if (n == 205)
		return'%';
	else if (n == 211)
		return'"';
	else if ((n >= 216) && (n <= 234))
		return' ';
	else if (n == 254)
		return' ';
	else if (n == 255)
		return' ';
	else if (n == 0)
		return'\0';
	else
		return'?';

	return 0;
}

/*************************************/
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDR_FILE_EXIT:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		case IDR_FILE_OPEN:
			openROM();
			break;
		case IDR_FILE_SAVE:
			saveROM();
			break;
		case IDR_OPTIONS_GRID:
			gridDialog.show();
			getSelectedTab()->invalidateTab();
			break;
		case IDR_OPTIONS_CHANGEBACK:
			changeBackground();
			getSelectedTab()->invalidateTab();
			break;
		case IDR_FILE_RECENT1:
		case IDR_FILE_RECENT2:
		case IDR_FILE_RECENT3:
		case IDR_FILE_RECENT4:
			openROM(LOWORD(wParam) - IDR_FILE_RECENT1);
			break;
		case IDR_IMAGE_IMPORT:
			getSelectedTab()->import();
			break;
		case IDR_IMAGE_EXPORT:
			getSelectedTab()->exportImage();
			break;
		case IDR_IMAGE_FLIP_HORIZONTAL:
			getSelectedTab()->flipImage(0);
			break;
		case IDR_IMAGE_FLIP_VERTICAL:
			getSelectedTab()->flipImage(1);
			break;
		case IDR_IMAGE_CLEAR:
			getSelectedTab()->clearImage();
			break;
		}
		break;
	case WM_NOTIFY:
		if (((LPNMHDR)lParam)->code == TCN_SELCHANGE)
			switchToTab(TabCtrl_GetCurSel(hMainTab));
		break;
	case WM_CLOSE:
		if (confirmClose() != 1)
			DestroyWindow(hWnd);
		break;
	case WM_DESTROY:
		options.saveConfig();
		PostQuitMessage(0);
		return 0;
		break;
	default:
		return DefWindowProc(hwnd, message, wParam, lParam);
	}
	return 0;
}

/*************************************/

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{

	MSG msg;
	hInst = hInstance;
	HACCEL hAccel;
	int handled;

	LoadString(hInst, IDS_TITLE, ff3seTitle, 80);

	if (initWindow() != 0)
		return 1;

	initProgram();

	if (GetDeviceCaps(screenHdc, BITSPIXEL) == 8)
		MessageBox(NULL, "Running this program in 256 color mode is not recommended",
			ff3seTitle, MB_OK | MB_ICONWARNING | MB_TASKMODAL);

	ShowWindow(hWnd, nCmdShow);

	if (InitTab() != 0)
		return 1;

	if (!(hAccel = LoadAccelerators(hInst, MAKEINTRESOURCE(IDR_ACCELERATOR))))
	{
		MessageBox(NULL, "Error initliazing program", ff3seTitle, MB_ICONERROR | MB_OK | MB_TASKMODAL);
		return 1;
	}

	/*LARGE_INTEGER t1,t2;
	int z;

	QueryPerformanceCounter(&t1);

	Bitmap bmp;
	for(z = 0;z < 100;z++)
		bmp.loadFile("c:/mine/bitmap/testbmps/4bit.pcx");

	QueryPerformanceCounter(&t2);

	FILE *file;

	file = fopen("time.txt","w+");
	fprintf(file,"%d\n",t2.QuadPart - t1.QuadPart);
	fclose(file);*/

	while (GetMessage(&msg, NULL, 0, 0))
	{
		handled = false;
		handled = TranslateAccelerator(hWnd, hAccel, &msg);
		if (!handled)
			handled = IsDialogMessage(getSelectedTab()->getHandle(), &msg);
		if (!handled)
			handled = IsDialogMessage(hDlg, &msg);
		if (!handled)
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	cleanup();

	return 0;
}

